# OB-USP-Agent
OBUSPA is fork of [BBFs obuspa](README.md) project with vendor integration of [uspd](https://dev.iopsys.eu/iopsys/uspd).

## Project Components

Project consists of following components:
- Application itself written in C programming language
- Documentation in a Markdown format
- Module tests usage bash scripts and python

## Build Instructions

obsupa is written using C programming language and depends on a number of components found in `OpenWrt` for building and running. It is advised to use iopsys feeds to resolve the dependencies.

## UCI Config

OBUSPA requires a configuration file to provide input parameters. It basically has uci options defined in global section which either configures the command line options for OBUSPA or non-datamodel inputs, rest of sections are generally a way to define parameters for below datamodel objects
```
Device.LocalAgent.
Device.MQTT.Client.
Device.STOMP.Connection.
```
Below is an example uci, complete list of supported uci parameters is available [link](./docs/api/obuspa.md), which is generated from [uci schema](./schemas/uci/obuspa.json)
```shellscript
config obuspa 'global'
	option enabled '1'
	option interface 'wan'
	#option ifname 'eth0.1'
	option debug '0'
	option log_level '1'
	#option db_file '/etc/obuspa/usp.db'
	#option role_file '/etc/obuspa/roles.json'
	option dm_caching_exclude '/etc/obuspa/dmcaching_exclude.json'
	#option trust_cert '/etc/obuspa/rootCA.pem'
	#option client_cert '/etc/obuspa/client.pem'
	#option log_dest '/var/log/obuspa'

config localagent 'localagent'
	option Enable '1'

config controller 'localcontroller'
	option EndpointID 'proto::interop-usp-controller'
	option Protocol 'MQTT'
	option Topic '/usp/controller'
	option mqtt 'localmqtt'
	option assigned_role_name 'full_access'

config mtp
	option Protocol 'MQTT'
	option ResponseTopicConfigured '/usp/endpoint'
	option mqtt 'localmqtt'

config mqtt 'localmqtt'
	option BrokerAddress '127.0.0.1'
	option BrokerPort '1883'
	option TransportProtocol 'TCP/IP'
```

## Concepts and Workflow
OBUSPA in itself has support for below datamodel objects
- Device.LocalAgent.
- Device.Security.
- Device.Time.
- Device.STOMP.
- Device.MQTT.
- Device.BulkData.

For rest of datamodel objects it depends on vendor integrations. Which mostly make the whole USP solution a monolithic block. To make it modular, we did the integration which gets the data over ubus(default rpc bus in openwrt), so that it can easily integrate with any 3rd party data-store. Currently it usages uspd as a data-store which exposes the datamodel object over ubus.

### Theory of operation
OBUSPA operations divided in three blocks {Init, Start, Stop}, vendor integration also follows the same.
- Vendor_Init
	- Reads mandatory objects from uspd, list of parameters available in (uspd_cache_mandatory_data)
	- Gets schema from uspd and register each parameter based on type
	- Registers below vendor hooks
		- Reboot
		- Factory Reset
    	- Transaction start
    	- Transaction commit
    	- Transaction abort
    	- ControllerTrust roles registration

- Vendor_Start
	- Calculate and update cached time for multi-instance objects
	- Loads objects in memory for which caching is disabled
	- Starts a ubus thread
    	- Responsible for getting the data exposed over ubus
    	- Registered for `usp.event` ubus event to signal generated event
    	- Registered for `usp.AddObj` ubus event to signal newly added object instances
    	- Registered for `usp.DelObj` ubus event to signal deleted object instances
	- Setup a socket pair to interact with Ubus thread

- Vendor_Stop
	- Free-up the dynamic allocated data and return

## How-to guide
### How to add controller trust roles
Currently OBUSPA does not support adding a new Role in the Role table, they are read-only and needed to be pre-created during init process. By default it support two roles, named
- full_access -- Having complete access of Device.
- Untrusted -- Having limited access, get of Device.DeviceInfo., some parameter of 'Device.LocalAgent.ControllerTrust.Challenge.' and operate of RequestChallenge/ChallengeResponse

By using vendor-hooks we have added support to add more roles by defining them in a json file(example [json](./test/files/etc/obuspa/roles.json)) and then configure it in 'obuspa.global.role_file' uci option.

### How to override the default role definitions
Default roles (full_access, Untrusted) can be overridden by defining a role in json with same name. This [example json](./test/files/etc/obuspa/roles.json) has overridden Untrusted role.

### How to change the max caching interval for multi-instance objects
Max caching internal is controlled by `MAX_CACHE_TIME` in `vendor_uspd.c`. Max caching interval is calculated at the runtime, it usually

```
max_cache_tme = MAX_CACHE_TIME + Time to get Device. from uspd
```

### How to optimize the group calls
OBUSPA provides mechanism to define the groups, so that the IPC cost can be minimised with Vendor integration. Currently the groups are created based on `MAX_GROUP_SEP(2)` meaning, objects are grouped with upto two delim(.) points, meaning, we have groups for Device.IP., Device.WiFi., Device.Ethernet. etc.

Now each call to a parameter result into calling the complete group, which is again not very efficient. So to optimize this, 'MIN_NUM_TO_GROUP(10)' is used.
So that, if the number of parameters are less than 'MIN_NUM_TO_GROUP', then only these parameters are get otherwise complete group is fetched from uspd and then required parameters are set accordingly.

### How to configure the ubus timeouts
All calls except operates usages `ubus_invoke` which requires a timeout value, which is handled by `USPD_TIMEOUT(10s)` in `vendor_uspd.c`.

### How does obuspa registers for new schema
In OBUSPA, registering new schema at runtime (after Start call) is not allowed, so to update the schema procd hooks used.
uspd pools for schema updates, once it detects schema change, it sends an ubus notification on usp.raw with '{"action":"schema_update_available"}' which reloads the obuspa services and thus it registers the new schema.

### How to define a multi-instance object in no caching list
It is supported to define the list of object for which obuspa shall not cache the instances using a json, meaning get value shall result in instance updates and then get_value from lower layer. Below is an example json:
```json
{
  "dmcaching_exclude": [
    "Device.DeviceInfo.ProcessStatus.Process.",
    "Device.Hosts.Host."
  ]
}
```

## Dependencies
OBUSPA has some compile time dependency and some run time dependencies as listed below:

### Build-Time Dependencies

To successfully build OBUSPA, following libraries are needed:

| Dependency   |                    Link                     |    License        |
| ------------ | ------------------------------------------- | ----------------- |
| libuci       | https://git.openwrt.org/project/uci.git     | LGPL 2.1          |
| libubox      | https://git.openwrt.org/project/libubox.git | BSD               |
| libubus      | https://git.openwrt.org/project/ubus.git    | LGPL 2.1          |
| libopenssl   | http://www.openssl.org                      | Apache License 2.0|
| libcurl      | https://github.com/curl/curl.git            | MIT               |
| libsqlite3   | https://www.sqlite.org/2020/                | PUBLICDOMAIN      |
| libmosquitto | https://mosquitto.org/files/source/         | EPL-2.0           |
| uspd         | https://dev.iopsys.eu/iopsys/uspd.git       | LGPL 2.1          |


### Run-Time Dependencies

In order to run the OBUSPA, following dependencies are needed to be running/available before OBUSPA.

| Dependency |                   Link                   | License  |
| ---------- | ---------------------------------------- | -------- |
| ubusd      | https://git.openwrt.org/project/ubus.git | LGPL 2.1 |
| uspd       | https://dev.iopsys.eu/iopsys/uspd.git    | LGPL 2.1 |

`uspd` is used to expose the datamodel object over `ubus`, it is the datastore for all datamodel objects except Device.LocalAgent.
